/*
 * Jsonix is a JavaScript library which allows you to convert between XML
 * and JavaScript object structures.
 *
 * Copyright (C) 2010  Aleksei Valikov, Highsource.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Jsonix.Model.ClassInfo = Jsonix.Class({
	name : null,
	baseTypeInfo : null,
	properties : null,
	structure : null,
	initialize : function(options) {
		Jsonix.Util.Ensure.ensureObject(options);
		Jsonix.Util.Ensure.ensureString(options.name);
		this.name = options.name;
		if (Jsonix.Util.Type.exists(options.baseTypeInfo)) {
			Jsonix.Util.Ensure.ensureObject(options.baseTypeInfo);
			this.baseTypeInfo = options.baseTypeInfo;
		}
		this.properties = [];
		if (Jsonix.Util.Type.exists(options.properties)) {
			Jsonix.Util.Ensure.ensureArray(options.properties);
			for ( var index = 0; index < options.properties.length; index++) {
				this.properties[index] = options.properties[index];
			}
		}
	},
	destroy : function() {
	},
	build : function(context) {
		if (this.structure !== null) {
			return;
		}
		var structure = {
			elements : null,
			attributes : {},
			anyAttribute : null,
			value : null,
			any : null
		};
		if (Jsonix.Util.Type.exists(this.baseTypeInfo)) {
			this.baseTypeInfo.buildStructure(context, structure);
		}
		this.buildStructure(context, structure);
		this.structure = structure;
	},
	buildStructure : function(context, structure) {
		for ( var index = 0; index < this.properties.length; index++) {
			var propertyInfo = this.properties[index];
			propertyInfo.buildStructure(context, structure);
		}
	},
	unmarshal : function(context, input) {
		this.build(context);
		var result = {};

		if (input.eventType !== 1) {
			throw "Parser must be on START_ELEMENT to read a class info.";
		}

		// Read attributes
		if (Jsonix.Util.Type.exists(this.structure.attributes)) {
			var attributeCount = input.getAttributeCount();
			if (attributeCount !== 0) {
				for ( var index = 0; index < attributeCount; index++) {
					var attributeName = input.getAttributeName(index);
					var attributeKey = attributeName.key;
					if (Jsonix.Util.Type.exists(this.structure.attributes[attributeKey])) {

						var attributePropertyInfo = this.structure.attributes[attributeKey];
						this.unmarshalProperty(context, input, attributePropertyInfo, result);
					}
				}
			}
		}
		// Read any attribute
		if (Jsonix.Util.Type.exists(this.structure.anyAttribute)) {
			var propertyInfo = this.structure.anyAttribute;
			this.unmarshalProperty(context, input, propertyInfo, result);
		}
		// Read elements
		if (Jsonix.Util.Type.exists(this.structure.elements)) {

			var et = input.next();
			while (et !== Jsonix.XML.Input.END_ELEMENT) {
				if (et === Jsonix.XML.Input.START_ELEMENT) {
					// New sub-element starts
					var elementName = input.getName();
					var elementKey = elementName.key;
					if (Jsonix.Util.Type.exists(this.structure.elements[elementKey])) {
						var elementPropertyInfo = this.structure.elements[elementKey];
						this.unmarshalProperty(context, input, elementPropertyInfo, result);
					} else if (Jsonix.Util.Type.exists(this.structure.any)) {
						// TODO Refactor

						var anyPropertyInfo = this.structure.any;
						this.unmarshalProperty(context, input, anyPropertyInfo, result);
					} else {
						// TODO report a validation error that element
						// is not expected
						throw 'Unexpected element [' + elementKey + '].';
					}
				} else if ((et === Jsonix.XML.Input.CHARACTERS || et === Jsonix.XML.Input.CDATA || et === Jsonix.XML.Input.ENTITY_REFERENCE) && Jsonix.Util.Type.exists(this.structure.mixed)) {
					// Characters and structure has a mixed property
					var mixedPropertyInfo = this.structure.mixed;
					this.unmarshalProperty(context, input, mixedPropertyInfo, result);
				} else if (et === Jsonix.XML.Input.SPACE || et === Jsonix.XML.Input.COMMENT || et === Jsonix.XML.Input.PROCESSING_INSTRUCTION) {
					// Ignore
				} else {
					throw "Illegal state: unexpected event type [" + et + "].";
				}
				et = input.next();
			}
		} else if (Jsonix.Util.Type.exists(this.structure.value)) {
			var valuePropertyInfo = this.structure.value;
			this.unmarshalProperty(context, input, valuePropertyInfo, result);
		} else {
			// Just skip everything
			input.nextTag();
		}
		if (input.eventType !== 2) {
			throw "Illegal state: must be END_ELEMENT.";
		}
		return result;
	},
	unmarshalProperty : function(context, input, propertyInfo, result) {
		var propertyValue = propertyInfo.unmarshal(context, this, input);
		if (Jsonix.Util.Type.exists(propertyValue)) {
			if (propertyInfo.collection) {
				if (!Jsonix.Util.Type.exists(result[propertyInfo.name])) {
					result[propertyInfo.name] = [];
				}
				for ( var index = 0; index < propertyValue.length; index++) {
					result[propertyInfo.name].push(propertyValue[index]);
				}

			} else {
				result[propertyInfo.name] = propertyValue;
			}
		}
	},
	marshal : function(context, value, output) {
		// TODO This must be reworked
		if (Jsonix.Util.Type.exists(this.baseTypeInfo)) {
			this.baseTypeInfo.marshal(context, value, output);
		}
		for ( var index = 0; index < this.properties.length; index++) {
			var propertyInfo = this.properties[index];
			var propertyValue = value[propertyInfo.name];
			if (Jsonix.Util.Type.exists(propertyValue)) {
				propertyInfo.marshal(context, this, propertyValue, output);
			}
		}
	},
	isInstance : function(value) {
		return Jsonix.Util.Type.isObject(value) && Jsonix.Util.Type.isString(value.TYPE_NAME) && value.TYPE_NAME === this.name;
	}
});